from __future__ import division
import numpy as np
import re
from collections import OrderedDict
import os

class TileUtils:
    def search_dir(self, directory):
        for dp, dn, fn in os.walk(os.path.expanduser(directory)):
            for f in fn:
                yield os.path.join(dp, f)

    def get_tiles(self, filenames, template, separator):
        """
        Given a list of tar pathnames + templates, parse Z X Ys
        """
        validtile = re.compile(template)
        for f in filenames:
            if validtile.match(f):
                matchstring = r"\d+%s\d+%s\d+" % (separator, separator)
                yield [
                    int(i) for i in re.compile(matchstring).findall(f)[-1].split(separator)
                ]

    def select_tiles(self, tiles, zoom):
        """
        Get the list of tiles for a zoom and their min / max Xs / Ys
        """
        tileCount, depth = tiles.shape

        if tileCount <= 0 or depth != 3:
            raise ValueError("Tile array must have shape of n > 0, 3 (recieved %s, %s)" % (tileCount, depth))

        tiles = tiles[np.where(tiles[:, 0] == zoom)]

        tileCount, depth = tiles.shape

        if tileCount <= 0 or depth != 3:
            raise ValueError("Tile array must have shape of n > 0, 3 (recieved %s, %s)" % (tileCount, depth))

        return tiles, tiles[:, 1].min(), tiles[:, 2].min(), tiles[:, 1].max(), tiles[:, 2].max()

    def get_super_tiles(self, subTiles, zoom):
        """
        Given an array of [z, x, y] tiles, find all their parent tiles at a particular zoom
        """
        if np.any(subTiles[:, 0] < zoom):
            raise ValueError("Cannot get super tiles of tile array w/ smaller zoom")
        zoomdiffs = 2 ** (subTiles[:, 0] - zoom)
        superTiles = subTiles // np.vstack(zoomdiffs)
        superTiles[:,0] = zoom

        return superTiles

    def get_unique_tiles(self, data):
        """
        Given an array of [z, x, y] tiles, find all the unique tiles
        """
        ncols = data.shape[1]
        dtype = data.dtype.descr * ncols
        struct = data.view(dtype)

        uniq = np.unique(struct)
        uniq = uniq.view(data.dtype).reshape(-1, ncols)
        return uniq


    def get_zoom_tiles(self, subTiles, superTiles, tile, tilefloor=16):
        """
        Given:
        (1) an array of [z, x, y] sub tiles at multiple zooms;
        (2) as array of [z, x, y] super tiles at one zoom, one for each record in the subtile array;
        (3) a single [z, x, y] from the above super tile array,
        find the maximum zoom level of tiles within that tile, and
        the maximum zoom level of complete coverage for that tile.
        If the same, returns <array of zMax tiles>, False
        If different, returns <array of zMax tiles>, <array of zMaxCov tiles>
        """

        if subTiles.shape != superTiles.shape:
            raise ValueError("Input sub and super tiles must have the same shape")

        subTiles = subTiles[np.all(superTiles == tile, axis=1)]
        subTileMin, subTileMax = subTiles[:, 0].min(), subTiles[:, 0].max()



        if subTileMax < tilefloor:
            raise ValueError("No tiles found below that floor")

        for zMaxCov in range(subTileMax, max([subTileMin - 1, tilefloor - 1]), -1):
            if 4 ** (zMaxCov - tile[0]) == np.where(subTiles[:, 0] == zMaxCov)[0].shape[0]:
                break


        if subTileMax != zMaxCov:
            return subTiles[np.where(subTiles[:, 0] == subTileMax)], subTiles[np.where(subTiles[:, 0] == zMaxCov)]
        else:
            return subTiles[np.where(subTiles[:, 0] == subTileMax)], False

    def get_sub_tiles(self, subTiles, superTiles):
        for t in self.get_unique_tiles(superTiles):
            z, x, y = t
            zMaxTiles, maxCovTiles = self.get_zoom_tiles(subTiles, superTiles, t)
            
            if not np.any(maxCovTiles):
                zMaxCov = False
            else:
                zMaxCov = maxCovTiles[0][0]

            yield OrderedDict({
                'zMaxTiles': zMaxTiles,
                'maxCovTiles': maxCovTiles,
                'zMax': zMaxTiles[0][0],
                'zMaxCov': zMaxCov,
                'x': x,
                'y': y,
                'z': z
            })

    def filter_tiles(self, tiles, zoomfloor):
        return tiles[np.where(tiles[:, 0] <= zoomfloor)]

    def get_fill_super_tiles(self, superTiles, fillTiles, fillThresh):
        for ct in ((np.all(superTiles == a, axis=1).sum(), a) for a in fillTiles):
            if ct[0] != fillThresh or ct[0] == 0:
                yield ct[1]

    def get_sub_base_zoom(self, px, py, pz, z):
        if z < pz:
            raise ValueError("cannot get base zoom of %s, is less than %s" % (z, pz))
        mult = 2 ** (z - pz)
        return (px * mult, py * mult)


def parse_template(template):
    """
    Parse and verify a pathname template
    """
    pattern = re.compile(r".*{z}.*{x}.*{y}.*(png|jpg|tif)")
    match = pattern.match(template)
    if pattern.match(template):    
        valPattern = re.compile(r"{(z|x|y)}")
        filepath = re.compile(r"(jpg|png|tif)$")
        sepmatch = re.compile(r"(?:{z})(/|-)(?:{x})(/|-)(?:{y})")
        separator = sepmatch.findall(template)[0]

        if len(separator) != 2 or separator[0] != separator[1]:
            raise ValueError('Too many / not matching separators!')
    
        return valPattern.sub('\d+', template), valPattern.sub('%s', template), separator[0]
    else:
        raise ValueError('Invalid template "%s"' % (template))


if __name__ == "__main__":
    TileUtils()
    parse_template()



#!/usr/bin/env python

import os, re

import click, mercantile, json
from rasterio.rio.options import creation_options

import untiler

from untiler.scripts.mbtiles_extract import MBTileExtractor

@click.group()
def cli():
    pass

@click.command()
@click.argument('input_dir', type=click.Path(exists=True))
@click.argument('output_dir', type=click.Path(exists=True))
@creation_options
@click.option('--compositezoom', '-c', default=13, type=int,
    help='Tile size to mosaic into [default=13]')
@click.option('--maxzoom', '-z', default=None, type=int,
    help='Force a maxzom [default=max in each compositezoom area]')
@click.option('--logdir', '-l', default=None, help="Location for log files [default=None]")
@click.option('--readtemplate', '-t', default="jpg/{z}/{x}/{y}.jpg", help="File path template [default='jpg/{z}/{x}/{y}.jpg']")
@click.option('--scenetemplate', '-s', default="{z}-{x}-{y}-tile.tif", help="Template for output scenetif filenames [default='{z}-{x}-{y}-tile.tif']")
@click.option('--workers', '-w', default=4, help="Number of workers in the processing pool [default=4]")
@click.option('--no-fill', '-x', is_flag=True, help="Don't fill in with lower zooms")
def streamdir(input_dir, output_dir, compositezoom, maxzoom, logdir, readtemplate, scenetemplate, workers, creation_options, no_fill):
    # with MBTileExtractor(input_dir) as mbtmp:
    #     print mbtmp.extract()
    untiler.stream_dir(input_dir, output_dir, compositezoom, maxzoom, logdir, readtemplate, scenetemplate, workers, creation_options, no_fill)

cli.add_command(streamdir)

@click.command()
@click.argument('mbtiles', type=click.Path(exists=True))
@click.argument('output_dir', type=click.Path(exists=True))
@creation_options
@click.option('--compositezoom', '-c', default=13, type=int,
    help='Tile size to mosaic into [default=13]')
@click.option('--maxzoom', '-z', default=None, type=int,
    help='Force a maxzom [default=max in each compositezoom area]')
@click.option('--scenetemplate', '-s', default="{z}-{x}-{y}-tile.tif", help="Template for output scenetif filenames [default='{z}-{x}-{y}-tile.tif']")
@click.option('--workers', '-w', default=4, help="Number of workers in the processing pool [default=4]")
@click.option('--no-fill', '-x', is_flag=True, help="Don't fill in with lower zooms")
def streammbtiles(mbtiles, output_dir, compositezoom, maxzoom, creation_options, scenetemplate, workers, no_fill):
    with MBTileExtractor(mbtiles) as mbtmp:
        input_tile_dir = mbtmp.extract()
        with open(os.path.join(input_tile_dir, 'tiles', 'metadata.json')) as ofile:
            metadata = json.loads(ofile.read())
        readtemplate = "tiles/{z}/{x}/{y}.%s" % (metadata['format'])
        untiler.stream_dir(input_tile_dir, output_dir, compositezoom, maxzoom, None, readtemplate, scenetemplate, workers, creation_options, no_fill)

cli.add_command(streammbtiles)

@click.command()
@click.argument('input_dir', type=click.Path(exists=True))
@click.option('--zoom', '-z', default=None, type=int,
    help='Zoom to inspect [default = all]')
@click.option('--readtemplate', '-t', default="jpg/{z}/{x}/{y}.jpg",
    help="File path template [default='jpg/{z}/{x}/{y}.jpg']")
def inspectdir(input_dir, zoom, readtemplate):
    untiler.inspect_dir(input_dir, zoom, readtemplate)

cli.add_command(inspectdir)

if __name__ == "__main__":
    cli()



#!/usr/bin/env python
from __future__ import with_statement
from __future__ import print_function
from __future__ import division
import os
from multiprocessing import Pool

import click
import mercantile as merc
import numpy as np
import rasterio
from rasterio import Affine
from rasterio.warp import reproject
try:
    from rasterio.warp import RESAMPLING as Resampling  # pre-1.0
except ImportError:
    from rasterio.warp import Resampling

import untiler.scripts.tile_utils as tile_utils


def make_affine(height, width, ul, lr):
    """
    Create an affine for a tile of a given size
    """
    xCell = (ul[0] - lr[0]) / width
    yCell = (ul[1] - lr[1]) / height
    return Affine(-xCell, 0.0, ul[0],
        0.0, -yCell, ul[1])


def affaux(up):
    return Affine(1, 0, 0, 0, -1, 0), Affine(up, 0, 0, 0, -up, 0)


def upsample(rgb, up, fr, to):
    up_rgb = np.empty((rgb.shape[0], rgb.shape[1] * up, rgb.shape[2] * up), dtype=rgb.dtype)

    reproject(
        rgb, up_rgb,
        src_transform=fr,
        dst_transform=to,
        src_crs="EPSG:3857",
        dst_crs="EPSG:3857",
        resampling=Resampling.bilinear)

    return up_rgb


def make_src_meta(bounds, size, creation_opts={}):
    """
    Create metadata for output tiles
    """

    ul = merc.xy(bounds.west, bounds.north)
    lr = merc.xy(bounds.east, bounds.south)

    aff = make_affine(size, size, ul, lr)

    ## default values
    src_meta = {
        'driver': 'GTiff',
        'height': size,
        'width': size,
        'count': 4,
        'dtype': np.uint8,
        'affine': aff,
        "crs": 'EPSG:3857',
        'compress': 'JPEG',
        'tiled': True,
        'blockxsize': 256,
        'blockysize': 256
    }

    for c in creation_opts.keys():
        src_meta[c] = creation_opts[c]

    return src_meta


def make_window(x, y, xmin, ymin, windowsize):
    """
    Create a window for writing a child tile to a parent output tif
    """
    if x < xmin or y < ymin:
        raise ValueError("Indices can't be smaller than origin")
    row = (y - ymin) * windowsize
    col = (x - xmin) * windowsize

    return (
            (row, row + windowsize),
            (col, col + windowsize)
        )


globalArgs = None


def make_image_array(imdata, outputSize):
    try:
        depth, width, height = imdata.shape

        if depth == 4:
            alpha = imdata[3]
        else:
            alpha = np.zeros((outputSize, outputSize), dtype=np.uint8) + 255

        return np.array([
            imdata[0 % depth, :, :],
            imdata[1 % depth, :, :],
            imdata[2 % depth, :, :],
            alpha
        ])
    except Exception as e:
        raise e


def load_image_data(imdata, outputSize):
    imsize, depth = imdata.shape

    if int(np.sqrt(imsize)) != outputSize:
        raise ValueError("Output size of %s ** 2 does not equal %s" % (outputSize, imsize))

    return imdata.reshape(outputSize, outputSize, depth).astype(np.uint8), imsize, depth


def global_setup(inputDir, args):
    global globalArgs
    globalArgs = args


def logwriter(openLogFile, writeObj):
    if openLogFile:
        print(writeObj, file=openLogFile)
        return


def streaming_tile_worker(data):
    size = 2 ** (data['zMax'] - globalArgs['compositezoom']) * globalArgs['tileResolution']
    out_meta = make_src_meta(merc.bounds(data['x'], data['y'], data['z']), size, globalArgs['creation_opts'])
    z, x, y = [int(i) for i in (data['z'], data['x'], data['y'])]
    filename = globalArgs['sceneTemplate'] % (z, x, y)
    subtiler = tile_utils.TileUtils()
    log = 'FILE: %s\n' % filename
    try:
        with rasterio.open(filename, 'w', **out_meta) as dst:
            if data['zMaxCov']:
                superTiles = subtiler.get_super_tiles(data['zMaxTiles'], data['zMaxCov'])

                fillbaseX, fillbaseY = subtiler.get_sub_base_zoom(data['x'], data['y'], data['z'], data['zMaxCov'])

                ## fill thresh == the number of sub tiles that would need to occur in a fill tile to not fill (eg completely covered)
                fThresh = 4 ** (data['zMax'] - data['zMaxCov'])

                fDiff = 2 ** (data['zMax'] - data['zMaxCov'])

                toFaux, frFaux = affaux(fDiff)

                if not globalArgs['no_fill']:
                    print('filling')
                    ## Read and write the fill tiles first
                    for t in subtiler.get_fill_super_tiles(superTiles, data['maxCovTiles'], fThresh):
                        z, x, y = [int(i) for i in t]
                        path = globalArgs['readTemplate'] % (z, x, y)
                        log += '%s %s %s\n' % (z, x, y)

                        with rasterio.open(path) as src:
                            imdata = src.read()

                        imdata = make_image_array(imdata, globalArgs['tileResolution'])

                        imdata = upsample(imdata, fDiff, frFaux, toFaux)

                        window = make_window(x, y, fillbaseX, fillbaseY, globalArgs['tileResolution'] * fDiff)
                        dst.write(imdata, window=window)


            baseX, baseY = subtiler.get_sub_base_zoom(data['x'], data['y'], data['z'], data['zMax'])

            for t in data['zMaxTiles']:
                z, x, y = [int(i) for i in t]
                path = globalArgs['readTemplate'] % (z, x, y)
                log += '%s %s %s\n' % (z, x, y)

                with rasterio.open(path) as src:
                    imdata = src.read()

                imdata = make_image_array(imdata, globalArgs['tileResolution'])

                window = make_window(x, y, baseX, baseY, globalArgs['tileResolution'])

                dst.write(imdata, window=window)
        if globalArgs['logdir']:
            with open(os.path.join(globalArgs['logdir'], '%s.log' % os.path.basename(filename)), 'w') as logger:
                logwriter(logger, log)

        return filename

    except Exception as e:
        click.echo("%s errored" % (path), err=True)
        raise e


def inspect_dir(inputDir, zoom, read_template):
    tiler = tile_utils.TileUtils()

    allFiles = tiler.search_dir(inputDir)

    template, readTemplate, separator = tile_utils.parse_template("%s/%s" % (inputDir, read_template))

    allTiles = np.array([i for i in tiler.get_tiles(allFiles, template, separator)])

    allTiles, _, _, _, _ = tiler.select_tiles(allTiles, zoom)

    for t in allTiles:
        z, x, y = t
        click.echo([x, y, z])


def stream_dir(inputDir, outputDir, compositezoom, maxzoom, logdir, read_template, scene_template, workers, creation_opts, no_fill):
    tiler = tile_utils.TileUtils()

    allFiles = tiler.search_dir(inputDir)

    template, readTemplate, separator = tile_utils.parse_template("%s/%s" % (inputDir, read_template))

    allTiles = np.array([i for i in tiler.get_tiles(allFiles, template, separator)])

    if allTiles.shape[0] == 0 or allTiles.shape[1] != 3:
        raise ValueError("No tiles were found for that template")

    if maxzoom:
        allTiles = tiler.filter_tiles(allTiles, maxzoom)

    if allTiles.shape[0] == 0:
        raise ValueError("No tiles were found below that maxzoom")

    _, sceneTemplate, _ = tile_utils.parse_template("%s/%s" % (outputDir, scene_template))

    pool = Pool(workers, global_setup, (inputDir, {
        'maxzoom': maxzoom,
        'readTemplate': readTemplate,
        'outputDir': outputDir,
        'tileResolution': 256,
        'compositezoom': compositezoom,
        'fileTemplate': '%s/%s_%s_%s_%s.tif',
        'sceneTemplate': sceneTemplate,
        'logdir': logdir,
        'creation_opts': creation_opts,
        'no_fill': no_fill
        }))

    superTiles = tiler.get_super_tiles(allTiles, compositezoom)

    for p in pool.imap_unordered(streaming_tile_worker, tiler.get_sub_tiles(allTiles, superTiles)):
        click.echo(p)

    pool.close()
    pool.join()


if __name__ == "__main__":
    stream_dir()
    inspect_dir()



import shutil, tempfile, os

import click

from mbutil import mbtiles_to_disk

import contextlib
import sys
from io import StringIO

@contextlib.contextmanager
def nostdout():
    save_stdout = sys.stdout
    sys.stdout = StringIO()
    yield
    sys.stdout = save_stdout

class MBTileExtractor:
    def __init__(self, path, subdir='tiles', scheme='xyz'):
        self.path = path
        self.tmpdir = tempfile.mkdtemp()
        self.subdir = subdir
        self.scheme = scheme
    def extract(self):
        with nostdout():
            mbtiles_to_disk(self.path, os.path.join(self.tmpdir, self.subdir), **{'scheme': self.scheme})
        return self.tmpdir
    def __enter__(self):
        return self
    def __exit__(self, ext_t, ext_v, trace):
        shutil.rmtree(self.tmpdir)
        if ext_t:
            click.echo("in __exit__")
